import os

from paste import urlmap
from paste import fileapp

from webob.exc import HTTPNotFound

from minreal.client import MinrealClient
from minreal.csp import CSPApp


class EchoClient(MinrealClient):
    """ A plugin that echos back any data sent it.

        This plugin also provides a simple HTML UI that allows user interaction
        as a demonstration.
    """

    @classmethod
    def app(cls):
        """ The 'app' method returns a WSGI app that wraps the plugin."""

        # Here, we build a PythonPaste URLMap WSGI application that will
        # that will dispatch to our various components.
        map = urlmap.URLMap(HTTPNotFound())

        # This plugin provides an index page,
        index_path = os.path.join(os.path.dirname(__file__),
                                  'static',
                                  'echo.html')
        map['/'] = fileapp.FileApp(index_path)

        # as well as an app to serve its static assets.
        static_path = os.path.join(os.path.dirname(__file__), 'static')
        map['/static'] = fileapp.DirectoryApp(static_path)

        # The CSPApp must be mounted somewhere in the WSGI tree as this is the
        # WSGI app that handles communication with the browser.
        map['/csp'] = CSPApp(cls)

        return map

    def __init__(self, send_func):
        # The plugin constructor takes one argument: a callable that sends its
        # (unicode) argument to the browser client.
        self._send = send_func

    def handle_data(self, chunk):
        """ Process a chunk of data transmitted from the browser to the plugin.

            This 'chunk' will be be a unicode string containing the transmitted
            data.
        """
        # The EchoClient plugin sends the chunk right back to the browser.
        self._send(chunk)



import os
import uuid
import warnings

import eventlet

import msgpack

from paste import urlmap
from paste import fileapp

from webob.exc import HTTPNotFound

from minreal.client import MinrealClient
from minreal.csp import CSPApp

DEFAULT_AVATAR = 'default_avatar.png'
DEFAULT_CENTER = (0, 0)
DEFAULT_SPEED = 10


def make_packet(type, payload):
    return dict(type=type, payload=payload)


class HalveLifeCharacter(object):

    @classmethod
    def create(cls, name):
        id_ = str(uuid.uuid4())
        return cls(id_, name)

    @classmethod
    def load(cls, id_, name, client, **specs):
        char = cls(id_, name)
        char.avatar = specs.get('avatar', DEFAULT_AVATAR)
        char.center = tuple(specs.get('center', DEFAULT_CENTER))
        char.destination = specs.get('destination')
        char.speed = specs.get('speed', DEFAULT_SPEED)
        return char

    def __init__(self, id_, name):
        self.id = id_
        self.name = name
        self.avatar = DEFAULT_AVATAR
        self.center = (0, 0)
        self.destination = None
        self.speed = DEFAULT_SPEED

    def update(self, update):
        for attr in ['avatar', 'center', 'destination', 'speed']:
            if attr in update:
                setattr(self, attr, update[attr])

    def serialize(self):
        attrs = ['id', 'name', 'avatar', 'center', 'destination', 'speed']
        return dict((attr, getattr(self, attr)) for attr in attrs)


class HalveLifeController(object):
    SEND_INTERVAL = 0.25

    def __init__(self):
        self._clients = []
        self._characters = {}
        self._packet_buffer = []

        eventlet.spawn_n(self._sender)

    def _sender(self):
        while True:
            if self._packet_buffer:
                batch = msgpack.packb(self._packet_buffer)
                self._packet_buffer = []
                clients_to_remove = []
                for client in self._clients:
                    try:
                        client.send(batch)
                    except Exception:
                        clients_to_remove.append(client)
                for client in clients_to_remove:
                    self._clients.remove(client)
            eventlet.sleep(self.SEND_INTERVAL)

    def _send_packet(self, type_, payload):
        self._packet_buffer.append(make_packet(type_, payload))

    def join(self, type_, name, destination, client):
        self._clients.append(client)
        char = HalveLifeCharacter.create(name)
        char.destination = tuple(destination)
        self._characters[char.id] = char
        char_attrs = char.serialize()
        print char_attrs
        self._send_packet('joined', char_attrs)

    def move(self, type_, id, center):
        char = self._characters[id]
        char.center = tuple(center)
        spec = {'id': char.id, 'center': char.center}
        self._send_packet('move', spec)

    def broadcast(self, type_, id, message):
        self._send_packet('broadcast', message)


class HalveLifeClient(MinrealClient):
    """ A plugin that echos back any data sent it.

        This plugin also provides a simple HTML UI that allows user interaction
        as a demonstration.
    """
    controller = HalveLifeController()

    @classmethod
    def app(cls):
        """ The 'app' method returns a WSGI app that wraps the plugin."""

        # Here, we build a PythonPaste URLMap WSGI application that will
        # that will dispatch to our various components.
        map = urlmap.URLMap(HTTPNotFound())

        # This plugin provides an index page,
        index_path = os.path.join(os.path.dirname(__file__),
                                  'static',
                                  'halvelife.html')
        map['/'] = fileapp.FileApp(index_path)

        # as well as an app to serve its static assets.
        static_path = os.path.join(os.path.dirname(__file__), 'static')
        map['/static'] = fileapp.DirectoryApp(static_path)

        # The CSPApp must be mounted somewhere in the WSGI tree as this is the
        # WSGI app that handles communication with the browser.
        map['/csp'] = CSPApp(cls)

        return map

    def __init__(self, send_func):
        # The plugin constructor takes one argument: a callable that sends its
        # (unicode) argument to the browser client.
        self.send = send_func

    def handle_data(self, encoded_batch):
        """ Process a chunk of data transmitted from the browser to the plugin.

            This 'chunk' will be be a unicode string containing the transmitted
            data.
        """
        packet = msgpack.unpackb(encoded_batch)
        print "***" + repr(packet) + "***"
        type_ = packet['type']
        handler = getattr(self.controller, type_, self._unknown_handler)
        if type_ == 'join':
            packet['payload']['client'] = self
        handler(type_, **packet['payload'])

    def _unknown_handler(self, type_, payload):
        message = "unknown handler requested: {}".format(type_)
        warnings.warn(message)



import base64
import copy
import json
import os
import time
import urllib
import uuid

import eventlet
from eventlet import queue

from paste import fileapp
from paste import urlmap

import webob
from webob.exc import HTTPNotFound

PACKET_ID = 0
PACKET_ENCODING = 1
PACKET_DATA = 2

PACKET_ENCODING_PLAIN = 0
PACKET_ENCODING_BASE64 = 1


class CSPResponseFactory(object):
    SEND_RESPONSE_FORMAT = (
        '{prebuffer}{preamble}'
        '{request_prefix}({data}){request_suffix}'
    )
    COMET_RESPONSE_FORMAT = (
        '{prebuffer}{preamble}'
        '{batch_prefix}({data}){batch_suffix}'
    )

    def render_handshake_response(self, session_vars, session_id):
        response_vars = {
            'data': json.dumps({'session': session_id}),
            'prebuffer': ' ' * session_vars['ps'],
            'preamble': session_vars['p'],
            'request_prefix': session_vars['rp'],
            'request_suffix': session_vars['rs'],
        }
        response = self.SEND_RESPONSE_FORMAT.format(**response_vars)
        headers = {'Content-type': session_vars['ct']}
        return headers, response

    def render_send_response(self, session_vars, message):
        headers = {'Content-type': session_vars['ct']}
        response_vars = {
            'data': json.dumps(message),
            'prebuffer': ' ' * session_vars['ps'],
            'preamble': session_vars['p'],
            'request_prefix': session_vars['rp'],
            'request_suffix': session_vars['rs'],
        }
        return headers, self.SEND_RESPONSE_FORMAT.format(**response_vars)

    def render_comet_response(self, session_vars, packets):
        headers = {'Content-type': session_vars['ct']}
        yield headers

        prebuffer = ' ' * session_vars['ps']
        preamble = session_vars['p']
        yield prebuffer + preamble

        for batch in packets:
            response_vars = {
                'data': json.dumps(batch),
                'prebuffer': ' ' * session_vars['ps'],
                'preamble': session_vars['p'],
                'batch_prefix': session_vars['bp'],
                'batch_suffix': session_vars['bs'],
            }
            yield self.COMET_RESPONSE_FORMAT.format(**response_vars)


class CSPSession(object):

    @staticmethod
    def parse_request_vars(request):
        parsed_request_vars = {}
        parsed_request_vars['s'] = request.get('s')
        parsed_request_vars['a'] = int(request.get('a', '-1'))
        parsed_request_vars['d'] = json.loads(request.get('d', '{}'))
        parsed_request_vars['n'] = bool(request.get('n'))
        return parsed_request_vars

    @staticmethod
    def parse_session_vars(request):
        parsed_session_vars = {}
        parsed_session_vars['rp'] = request.get('rp', '')
        parsed_session_vars['rs'] = request.get('rs', '')
        parsed_session_vars['du'] = int(request.get('du', '30'))
        parsed_session_vars['is'] = bool(int(request.get('is', '0')))
        parsed_session_vars['i'] = int(request.get('i', '0'))
        parsed_session_vars['ps'] = int(request.get('ps', '0'))
        parsed_session_vars['p'] = request.get('p', '')
        parsed_session_vars['bp'] = request.get('bp', '')
        parsed_session_vars['bs'] = request.get('bs', '')
        parsed_session_vars['g'] = bool(request.get('g', ''))
        parsed_session_vars['se'] = bool(request.get('se', ''))
        parsed_session_vars['ct'] = request.get('ct', 'text/html')
        return parsed_session_vars

    def __init__(self, session_id, client_factory, request_vars, session_vars):
        self._id = session_id
        self._vars = session_vars.copy()
        self._next_packet_id = 1
        self._packet_buffer = []

        self._packet_queue = queue.Queue()
        self._client = client_factory(self.add_chunk)
        self._client_gt = eventlet.spawn(self._client.run)

    @classmethod
    def create(cls, request_vars, session_vars, client_factory):
        session_uuid = str(uuid.uuid4())
        return cls(session_uuid, client_factory, request_vars, session_vars)

    @property
    def id(self):
        return self._id

    @property
    def session_vars(self):
        return self._vars.copy()

    def update_session_vars(self, new_session_vars):
        self._vars.update(new_session_vars)

    def add_chunk(self, chunk):
        self._packet_buffer.append(
            [self._next_packet_id, 1, base64.b64encode(chunk)]
        )
        self._next_packet_id += 1
        self._packet_queue.put(True)

    def _ack_packets(self, ack_id):
        buffer_copy = copy.copy(self._packet_buffer)
        self._packet_buffer = []
        for packet in buffer_copy:
            if packet[PACKET_ID] > ack_id:
                self._packet_buffer.append(packet)

    def _batch_packets(self):
        return [packet for packet in self._packet_buffer]

    def send(self, request_vars, session_vars):
        self._ack_packets(request_vars['a'])
        for packet in request_vars['d']:
            data = packet[PACKET_DATA]
            if packet[PACKET_ENCODING] == PACKET_ENCODING_BASE64:
                print repr(data)
                data = base64.b64decode(data)
            self._client.handle_data(data)
        return "OK"

    def comet(self, request_vars, session_vars):
        self._ack_packets(request_vars['a'])
        if not session_vars['is']:
            if self._packet_buffer:
                yield self._batch_packets()
            else:
                try:
                    self._packet_queue.get(timeout=session_vars['du'])
                except queue.Empty:
                    pass
                yield self._batch_packets()
        else:
            start = time.time()
            waited = 0.0
            try:
                timeout = session_vars['du']
                while self._packet_queue.get(timeout=timeout-waited):
                    yield self._batch_packets()
                    waited += time.time() - start
            except queue.Empty:
                pass
            yield self._batch_packets()


class CSPApp(object):

    def __init__(self, client_factory):
        self._client_factory = client_factory
        self._sessions = {}
        self._response_factory = CSPResponseFactory()

        self._wsgi_app = urlmap.URLMap(HTTPNotFound())
        self._wsgi_app['/handshake'] = self.handshake
        self._wsgi_app['/send'] = self.send
        self._wsgi_app['/comet'] = self.comet
        static_path = os.path.join(os.path.dirname(__file__), 'static')
        self._wsgi_app['/static'] = fileapp.DirectoryApp(static_path)

    def __call__(self, environ, start_response):
        return self._wsgi_app(environ, start_response)

    def handshake(self, environ, start_response):
        request = webob.Request(environ).params
        request_vars = CSPSession.parse_request_vars(request)
        session_vars = CSPSession.parse_session_vars(request)
        session = CSPSession.create(request_vars,
                                    session_vars,
                                    self._client_factory)
        self._sessions[session.id] = session

        headers, response = self._response_factory.render_handshake_response(
            session.session_vars, session.id
        )
        return webob.Response(headers=headers, app_iter=response)(
            environ, start_response)

    def send(self, environ, start_response):
        request = webob.Request(environ).params
        request_vars = CSPSession.parse_request_vars(request)
        session_vars = CSPSession.parse_session_vars(request)
        session = self._sessions.get(request_vars['s'])
        message = session.send(request_vars, session_vars)
        headers, response = self._response_factory.render_send_response(
            session.session_vars, message
        )
        return webob.Response(headers=headers, app_iter=response)(
            environ, start_response)

    def comet(self, environ, start_response):
        request = webob.Request(environ).params
        request_vars = CSPSession.parse_request_vars(request)
        session_vars = CSPSession.parse_session_vars(request)
        session = self._sessions.get(request_vars['s'])
        packets = session.comet(request_vars, session_vars)

        response = self._response_factory.render_comet_response(
            session.session_vars, packets
        )
        headers = response.next()
        return webob.Response(headers=headers, app_iter=response)(
            environ, start_response)



import os
import uuid

from paste import urlmap
from paste import fileapp

from webob.exc import HTTPNotFound

from minreal.client import MinrealClient
from minreal.csp import CSPApp


class TicTacToeController(object):

    def __init__(self):
        self._wait_queue = []
        self._games = {}

    def join(self, player):
        player_id, player_send = player
        if self._wait_queue and player not in self._wait_queue:
            opponent = self._wait_queue.pop(0)
            opponent_id, opponent_send = opponent
            game_id = str(uuid.uuid4())
            self._games[game_id] = TicTacToeGame(player, opponent)
            player_send('game:{}:{}'.format(game_id, 'X'))
            opponent_send('game:{}:{}'.format(game_id, 'O'))
        else:
            self._wait_queue.append(player)

    def play(self, game_id, mark, square):
        square = int(square)
        if mark == 'X':
            winner = self._games[game_id].mark_X(square)
        else:
            winner = self._games[game_id].mark_O(square)
        if winner is not False:
            del self._games[game_id]


class TicTacToeGame(object):
    WINNERS = (
        (0, 1, 2), (3, 4, 5), (6, 7, 8),  # Vertical
        (0, 3, 6), (1, 4, 7), (2, 5, 8),  # Horizontal
        (0, 4, 8), (2, 4, 6),             # Diagonal
    )

    def __init__(self, playerX, playerO):
        self.playerX = playerX
        self.playerO = playerO
        self._nextPlayer = playerX

        self._squares = [None] * 9

    def _send_mark(self, mark, square):
        message = 'mark:{}:{}'.format(mark, square)
        self.playerX[1](message)
        self.playerO[1](message)

    def mark_X(self, square):
        if self._nextPlayer == self.playerX and not self._squares[square]:
            self._nextPlayer = self.playerO
            self._send_mark('X', square)
            self._squares[square] = 'X'
            return self.winner()

    def mark_O(self, square):
        if self._nextPlayer == self.playerO and not self._squares[square]:
            self._nextPlayer = self.playerX
            self._send_mark('O', square)
            self._squares[square] = 'O'
            return self.winner()

    def winner(self):
        for a, b, c in self.WINNERS:
            if (self._squares[a] and
                self._squares[b] and
                self._squares[c] and
                (self._squares[a] ==
                 self._squares[b] ==
                 self._squares[c])):
                winning_mark = self._squares[a]
                if winning_mark == 'X':
                    self.playerX[1]('result:winner')
                    self.playerO[1]('result:loser')
                else:
                    self.playerO[1]('result:winner')
                    self.playerX[1]('result:loser')
                return winning_mark
        if all(self._squares):
            self.playerO[1]('result:tie')
            self.playerX[1]('result:tie')
            return None
        return False


class TicTacToeClient(MinrealClient):
    """ A plugin that echos back any data sent it.

        This plugin also provides a simple HTML UI that allows user interaction
        as a demonstration.
    """
    controller = TicTacToeController()

    @classmethod
    def app(cls):
        """ The 'app' method returns a WSGI app that wraps the plugin."""

        # Here, we build a PythonPaste URLMap WSGI application that will
        # that will dispatch to our various components.
        map = urlmap.URLMap(HTTPNotFound())

        # This plugin provides an index page,
        index_path = os.path.join(os.path.dirname(__file__),
                                  'static',
                                  'tictactoe.html')
        map['/'] = fileapp.FileApp(index_path)

        # as well as an app to serve its static assets.
        static_path = os.path.join(os.path.dirname(__file__), 'static')
        map['/static'] = fileapp.DirectoryApp(static_path)

        # The CSPApp must be mounted somewhere in the WSGI tree as this is the
        # WSGI app that handles communication with the browser.
        map['/csp'] = CSPApp(cls)

        return map

    def __init__(self, send_func):
        # The plugin constructor takes one argument: a callable that sends its
        # (unicode) argument to the browser client.
        self.send = send_func

        self._uuid = str(uuid.uuid4())

    def handle_data(self, message):
        """ Process a chunk of data transmitted from the browser to the plugin.

            This 'chunk' will be be a unicode string containing the transmitted
            data.
        """
        print "***" + message + "***"
        if message == 'join':
            self.controller.join((self._uuid, self.send))
        elif message.startswith('play'):
            token, game_id, mark, square = message.split(':')
            self.controller.play(game_id, mark, square)



import eventlet
eventlet.monkey_patch()

import importlib

from argparse import ArgumentParser  # noqa: E402

from .server import MinrealServer  # noqa: E402


def main():
    parser = ArgumentParser()
    parser.add_argument('clients', metavar='CLIENTS', nargs='+')

    args = parser.parse_args()

    clients = {}
    for client_spec in args.clients:
        path, client_path = client_spec.split(':', 1)
        modulename, clientname = client_path.split(':', 1)
        clients[path] = getattr(importlib.import_module(modulename),
                                clientname)

    server = MinrealServer(clients)
    server.run()


if __name__ == '__main__':
    main()



import eventlet
from eventlet import wsgi

from .app import MinrealApp


class MinrealServer(object):

    def __init__(self, clients):
        self._clients = clients

    def run(self, host='0.0.0.0', port=5001):
        app = MinrealApp(self._clients)
        wsgi.server(eventlet.listen((host, port)), app)



from setuptools import setup, find_packages

version = '0.1'

DESCRIPTION = ("A Python comet server build on CSP (Comet Session"
               " Protocol) and providing a TCP socket proxy.")

LONG_DESCRIPTION = DESCRIPTION + """

## Running the examples
The standalone server can be run like:
```bash
$ mrl <mount path>:<client plugin module dotted name>:<client plugin class>
```

So you can run the examples like:
```bash
# For the echo server example
$ mrl echo:minreal.examples.echo:EchoClient

# For the TCPSocket IRC example
$ mrl tcp:minreal.examples.tcp:TCPClient
```

## Developing a plugin
Take a look at the [annotated example]
(https://github.com/desmaj/minreal/blob/master/minreal/examples/echo.py)
to get started.
"""

setup(name='minreal',
      version=version,
      description=DESCRIPTION,
      long_description=LONG_DESCRIPTION,
      classifiers=[],
      keywords='',
      author='Matthew Desmarais',
      author_email='matthew.desmarais@gmail.com',
      url='',
      license='MIT',
      packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          'paste',
          'eventlet',
          'webob',
          'msgpack-python',
      ],
      entry_points="""
      [console_scripts]
      mrl = minreal.run:main
      """,
      )



import os
import sys

import eventlet
from eventlet.green import socket

import msgpack

from paste import urlmap
from paste import fileapp

from webob.exc import HTTPNotFound

from minreal.client import MinrealClient
from minreal.csp import CSPApp


class TCPClient(MinrealClient):

    @classmethod
    def app(cls):
        map = urlmap.URLMap(HTTPNotFound)

        index_path = os.path.join(os.path.dirname(__file__),
                                  'static',
                                  'tcp.html')
        map['/'] = fileapp.FileApp(index_path)

        static_path = os.path.join(os.path.dirname(__file__), 'static')
        map['/static'] = fileapp.DirectoryApp(static_path)

        map['/csp'] = CSPApp(cls)

        return map

    def __init__(self, send_func):
        self._send = send_func
        self._sock = None

    def handle_data(self, message):
        # message = ''.join([chr(ord(c)) for c in message])
        payload = msgpack.unpackb(message)
        if 'connect' in payload:
            host, port_str = payload['connect'].split('::')
            port = int(port_str)
            self._sock = socket.socket()
            self._sock.connect((host, port))
            response = msgpack.packb({'type': 'STATUS',
                                      'value': 'OPEN'})
            self._send(response)
            eventlet.spawn(self._reader)
        elif 'close' in payload:
            self._sock.close()
            status = msgpack.packb({'type': 'STATUS',
                                    'value': 'CLOSED'})
            self._send(status)
        elif 'content' in payload:
            try:
                self._sock.send(payload['content'])
            except Exception as exc:
                print exc
                self._sock.close()
        else:
            message = 'Unknown payload type: {}\n'.format(payload.type)
            sys.write(message)

    def _reader(self):
        while True:
            chunk = self._sock.recv(2048)
            if not chunk:
                self._sock.close()
                break
            else:
                content = msgpack.packb({'type': 'CONTENT',
                                         'value': chunk})
                self._send(content)



import os

from paste import fileapp
from paste import urlmap

from webob.exc import HTTPNotFound


class MinrealApp(object):

    def __init__(self, clients):
        self._app = urlmap.URLMap(HTTPNotFound())
        for path, client_klass in clients.items():
            self._app['/' + path] = client_klass.app()
        static_path = os.path.join(os.path.dirname(__file__), 'static')
        self._app['/static'] = fileapp.DirectoryApp(static_path)

    def __call__(self, environ, start_response):
        return self._app(environ, start_response)
